// Pingus - A free Lemmings clone
// Copyright (C) 1999 Ingo Ruhnke <grumbel@gmx.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <vector>
#include <cstdio>
#include <ctime>

#include "PingusLevelDesc.hh"
#include "PinguTileMap.hh"
#include "PinguBmpMap.hh"
#include "PinguSpotMap.hh"
#include "PinguRandomMap.hh"

#include "PingusError.hh"
#include "algo.hh"
#include "PingusSound.hh"
#include "globals.hh"
#include "Background.hh"
#include "World.hh"
#include "traps/traps.hh"
#include "Result.hh"
#include "actions/pingu_actions.hh"

//CL_ResourceManager* World::resource;

World::World()
{
  released_pingus = 0;
  colmap = new ColMap;
  exit_world = false;
}

World::~World()
{
  // Bug: This crashes, if pingus is exited
  // delete map;
  // delete  bg;
}

void 
World::draw_pingus(int x_offset, int y_offset, double s=1.0)
{
  for(PinguIter pingu = pingus.begin(); pingu != pingus.end(); ++pingu) {
    if ((*pingu)->get_status() != dead) {
      (*pingu)->draw_offset(x_offset, y_offset, s);
    }
  }
}

// draw the basic background pattern
void 
World::draw_background(int x_offset, int y_offset, double s=1.0) const
{
  bg->draw_offset(x_offset, y_offset, s);

  for(unsigned int i=0; i < hotspot_bg.size(); ++i) 
    hotspot_bg[i]->draw_offset(x_offset, y_offset, s);
			       
}

void 
World::draw_playground(int x_offset, int y_offset, double s=1.0) const
{
  // Draw the gfx map
  if (map)
    map->draw_offset(x_offset, y_offset, s);

  // Draw the exits
  for(unsigned int i=0; i < exits.size(); ++i) {
    exits[i]->draw_offset(x_offset, y_offset, s);
  }
  
  // Draw the entrances
  for(unsigned int i=0; i < entrance.size(); ++i) {
    entrance[i]->draw_offset(x_offset, y_offset, s);
  }

  // Draw the hotspots
}


void 
World::draw_foreground(int x_offset, int y_offset, double s=1.0) const
{
  for(unsigned int i=0; i < traps.size(); ++i) 
    traps[i]->draw_offset(x_offset, y_offset, s);

  for(unsigned int i=0; i < hotspot_fg.size(); ++i) {
    hotspot_fg[i]->draw_offset(x_offset, y_offset, s);
  }
  // BUG: This give a warning while compiling, because the function
  // isn't const, but setting it const give some other problems.
  particle.draw_offset(x_offset, y_offset, s);
}

// Merge the different layers on the screen together
void 
World::draw_offset(int x, int y, double s = 1.0) const
{
  draw_background(x, y, s);
  draw_playground(x, y, s);
  draw_pingus(x, y, s);
  draw_foreground(x, y, s);
}

void 
World::let_move()
{
  if (allowed_pingus == released_pingus && pingus_out == 0) {
    cout << "World finished" << endl;
    exit_world = 200;
  }

  if (do_armageddon == true) {
    ++exit_world;
  }
  
  if (do_armageddon && armageddon_count != pingus.end()) {
    if (armageddon_count == 0) {
      armageddon_count = pingus.begin();
    } else {
      int i = 0;
      for(PinguIter pingu = armageddon_count; i < 5; ++pingu, ++i) {
	armageddon_count = pingu;	
	if (armageddon_count == pingus.end())
	  break;
	if ((*pingu)->get_status() != dead)
	  (*pingu)->set_action(new Bomber);
	else
	  --i;
      }
    }
  }
  
  // Create new pingus, if enough time is passed
  if (!do_armageddon) {
    for(unsigned int i=0; i < entrance.size(); ++i) {
      if (pingus.size() < allowed_pingus
	  && entrance[i]->pingu_ready()) {
	pingus.push_back(entrance[i]->get_pingu());
	++released_pingus;
      }
    }
  }
  
  // Let all pingus move and
  // Let the pingus catch each other and
  // Let the traps catch the pingus and
  // Let the exit catch the pingus
  pingus_out = pingus.size();

  for(PinguIter pingu = pingus.begin(); pingu != pingus.end(); ++pingu) {
    (*pingu)->let_move();
    
    if ((*pingu)->get_status() == dead
	|| (*pingu)->get_status() == exited) {
      --pingus_out;
      continue;
    }
    
    if ((*pingu)->need_catch()) {
      for(PinguIter j=pingus.begin(); j != pingus.end(); ++j) {
	(*pingu)->catch_pingu(*j);
      }
    }
    
    for(unsigned int j=0; j < traps.size(); ++j) 
      traps[j]->catch_pingu(*pingu);
      
    for(unsigned int j=0; j < exits.size(); ++j) 
      exits[j]->catch_pingu(*pingu);
      
    for(unsigned int i=0; i < traps.size(); ++i) 
      traps[i]->let_move();
  }
  particle.let_move();
}

void 
World::save()
{
  // Empty should save the state of the wolrd 
  // Maybe that should be implemented using a demo file
}

void 
World::load()
{
  // Empty should load the state of the world
  // Maybe that should be implemented using a demo file
}

void
World::load_data()
{
  //  resource = res;
}

void 
World::init(PLF* level_data, Result& r)
{
  PingusLevelDesc leveldesc(level_data);

  leveldesc.draw(PingusLevelDesc::LOADING);

  do_armageddon = false;
  armageddon_count = 0;
  exit_world = 0;
  result = &r;
      /*
    CL_Surface* tex;
    CL_Font* font = CL_Res_Font::load("Fonts/courier_h", *resource());
    CL_Font* font_s = CL_Res_Font::load("Fonts/courier", *resource());
    
    if (rand() % 2) {
      tex = CL_Res_Surface::load("Textures/rocktile", *resource());
    } else {
      tex = CL_Res_Surface::load("Textures/stone", *resource());
    }
    
    for(int y=0; y < CL_Display::get_height(); y += tex->get_height())
      for(int x=0; x < CL_Display::get_width(); x += tex->get_width())
	tex->put_screen(x, y);
    
    font_s->print_center(CL_Display::get_width() / 2 + 4, 
			 CL_Display::get_height() / 2 + 4,
			 "loading...");
    font->print_center(CL_Display::get_width() / 2, 
		       CL_Display::get_height() / 2,
		       "loading...");
    
    CL_Display::flip_display(true);
    */
  

  if (verbose)
    cout << "Wolrd init" << endl;

  // load the foreground map
  switch (level_data->map_type()) {
  case TILE:
    map = new PinguTileMap(level_data, resource());
    break;
  case BMP:
    map = new PinguBmpMap(level_data->get_fg(), resource());
    break;
  case SPOT:
    map = new PinguSpotMap(level_data->get_fg(), resource());
    break;
  case RANDOM:
    map = new PinguRandomMap;
    break;
  default:
    throw PingusError("Error in PLF file: Undef Maptype");
    break;
  }

  if (level_data->get_mapfile().type == ResDescriptor::AUTO) {
    if (verbose) cout << "Using AutoColMap" << endl;
    colmap = map->get_colmap();;
    assert(colmap);
  } else {
    if (verbose) cout << "Using StaticColMap" << endl;
    colmap->load(level_data->get_mapfile());
  }

  if (verbose)
    cout << "Loaded Colmap" << endl;

  if (verbose)
    cout << "Loaded pingu map" << endl;

  // load the background map
  bg = new Background(level_data->get_bg());

  if (verbose)
    cout << "Loaded Background" << endl;

  allowed_pingus = level_data->get_pingus();
  released_pingus = 0;
  pingus_out = 0;

  result->set_total(allowed_pingus);

  //  int pal;
  //  srand(time(NULL));

  // Creating Exit and Entrance
  vector<exit_data>     exit_d     = level_data->get_exit();
  vector<entrance_data> entrance_d = level_data->get_entrance();

  for(unsigned int i=0; i < exit_d.size(); ++i) 
    exits.push_back(new Exit(exit_d[i]));
  
  for(unsigned int i=0; i < entrance_d.size(); ++i) 
    entrance.push_back(new Entrance(entrance_d[i]));

  vector<trap_data>     trap_d     = level_data->get_traps();
  
  for(unsigned int i=0; i < trap_d.size(); ++i)
    traps.push_back(get_trap(trap_d[i]));

  // Creating the foreground and background hotspots
  vector<hotspot_t> hspot_fg = level_data->get_hotspot_fg();
  vector<hotspot_t> hspot_bg = level_data->get_hotspot_bg();

  for(unsigned int i=0; i < hspot_bg.size(); ++i)
    hotspot_bg.push_back(new Hotspot(hspot_bg[i], resource()));

  for(unsigned int i=0; i < hspot_fg.size(); ++i)
    hotspot_fg.push_back(new Hotspot(hspot_fg[i], resource()));
 
  // Setting Pingus at random positions
  //  pingu.resize(0);

  //  for(int i=0; i < pingu.size(); ++i) 
  //    pingu[i] = new Pingu(rand() % 800, rand() % 300);

  Pingu::set_colmap(colmap);
  Pingu::set_map(map);
  Pingu::SetParticleHolder(&particle);
  Trap::SetParticleHolder(&particle);

  if (verbose)
    cout << "Finished Word init" << endl;

  if (sound_enabled) PingusSound::play(PingusSound::LETSGO);
  
  leveldesc.draw(PingusLevelDesc::FINISHED);
  while(!CL_Mouse::left_pressed()) 
    CL_System::keep_alive();
  
  while(CL_Mouse::left_pressed())
    CL_System::keep_alive();
}

PinguHolder*
World::get_pingu_p()
{
  return &pingus;
}

int
World::get_width()
{
  assert(map);
  return map->get_width();  
}

int
World::get_height()
{
  assert(map);
  return map->get_height();
}

Result
World::write_results()
{
  int saved   = 0;
  int killed  = 0;
  int unknown = 0;

  result->set_total(allowed_pingus);

  for(PinguIter pingu=pingus.begin(); pingu != pingus.end(); ++pingu) {
    switch((*pingu)->get_status()) {
    case dead:
      ++killed;
      break;
    case exited:
      ++saved;
      break;
    default:
      ++unknown;
    }
  }

  result->set_unknown(unknown);
  result->set_saved(saved);
  result->set_killed(killed);
  
  return* result;
}

void 
World::armageddon(void)
{
  cout << "Doing armo" << endl;
  do_armageddon = true;
}

bool
World::is_finished(void)
{
  if (exit_world > 150) {
    return true;
  } else {
    return false;
  }
}

void
World::print_status(void)
{
  cout << "Released Pingus: " << released_pingus << endl;
  cout << "Allowed Pingus:  " << allowed_pingus << endl;
  cout << "Pingus Out:      " << pingus_out << endl;
}

/* EOF */
